from cytoolz.curried import assoc, compose

from genomoncology import kms
from genomoncology.parse.doctypes import DocType, __CHILD__, __TYPE__
from genomoncology.pipeline.transformers import register, name_mapping

NAME_MAPPING = {
    "hgvs_g": "hgvs_g",
    "functional_region__string": "functional_region",
    "aa_change__string": "aa_change",
    "count__int": "count",
    "total__int": "total",
    "frequency__float": "frequency",
    "csra": "csra",
}

register(
    input_type=DocType.MTDB_RECORD,
    output_type=DocType.MTDB,
    transformer=compose(
        lambda x: assoc(x, __TYPE__, DocType.MTDB.value),
        name_mapping(NAME_MAPPING),
        lambda x: assoc(x, "hgvs_g", kms.annotations.to_csra(x)),
    ),
)

register(
    input_type=DocType.MTDB_RECORD,
    output_type=DocType.MTDB,
    transformer=compose(lambda x: assoc(x, __CHILD__, DocType.MTDB.value)),
    is_header=True,
)



from datetime import datetime as dt


# prints a user friendly error message if the dob is not formatted properly
def check_dob_format(dob):
    try:
        if dob is not None:
            dt.strptime(dob, "%Y-%m-%d")
    except ValueError as error:
        error.args = (
            "The date `{0}` does not match the format YYYY-MM-DD".format(dob),
        )
        raise


def filter_del_lines_from_load_annotations(call):
    return call.get("alt", "") != "<DEL>"


def filter_bnd_cpx_ctx_lns_from_load_annotations(call):
    """If the dict['mutation_type'] is not equal to any of the unwanteds then
    return true so that the filter option will gather all of the lines that
    were considered true and dispose the lines that are false. Essentially we
    are leveraging the filter() function to remove any lines where
     'mutation_type'==unwanted"""
    unwanteds = ("BND", "CPX", "CTX")
    return call.get("mutation_type", "") not in unwanteds



from cytoolz.curried import curry, unique, concat
from glom import Coalesce

from genomoncology.kms.match import (
    calls_to_alterations,
    transform_match_response,
)
from genomoncology.parse import DocType, __TYPE__
from genomoncology.parse.ensures import ensure_collection


@curry
def match_trials(calls, diseases=None, sdk=None, dob=None, gender=None):
    alterations = calls_to_alterations(calls)
    diseases = ensure_collection(diseases or ["ANY"])

    response = sdk.call_with_retry(
        sdk.trials.match_trials_post,
        alterations=alterations,
        diseases=diseases,
        date_of_birth=dob,
        gender=gender,
    )

    return transform_match_response(response, DEFAULT_SPEC)


DEFAULT_SPEC = {
    __TYPE__: Coalesce(default=DocType.TRIAL.value),
    "nct_id": "nct_id",
    "title": "title",
    "phase": "phase",
    "recruiting_status": "recruiting_status",
    "trigger_alterations": (
        "detected_alterations",
        ["trigger_alterations"],
        lambda x: list(unique(concat(x))),
    ),
    "trigger_diseases": (
        "matched_diseases",
        ["matched_disease"],
        lambda x: list(unique(x)),
    ),
}



from cytoolz.curried import curry

from genomoncology.cli.const import GRCh37
from genomoncology.kms.annotations import convert_to_csra, to_csra
from genomoncology.parse import (
    is_call_or_variant,
    DocType,
    __CHILD__,
    __TYPE__,
)
from genomoncology.pipeline.converters import obj_to_dict
from gosdk.models import VariantInterpretationResponse


@curry
def get_variant_interpretations(
    data, sdk=None, template_name=None, delete_if_exists=False, build=GRCh37
):
    csra_batch = convert_to_csra(
        [{**d, "build": build} for d in data], add_build=False
    )

    variant_interpretations_response = (
        sdk.call_with_retry(
            sdk.variant_interpretations.generate_variant_interpretations,
            batch=csra_batch,
            template_names=[template_name],
            delete_if_exists=delete_if_exists,
            build=build,
        )
        if csra_batch
        else None
    )

    variant_interpretations_calls = []
    for call in data:
        variant_interpretations_calls.append(
            add_interpretations_to_calls(
                variant_interpretations_response, call
            )
        )

    return variant_interpretations_calls


def add_interpretations_to_calls(
    interpretations_response: VariantInterpretationResponse, call: dict
):
    if interpretations_response and is_call_or_variant(call):
        csra = to_csra(call, add_build=False)
        variant_interpretation_data = obj_to_dict(
            None, interpretations_response.get_var_int_data(csra) or {}
        )
        call["variant_interpretations"] = variant_interpretation_data.get(
            "variant_interpretations", []
        )
        call["protein_effects"] = variant_interpretation_data.get(
            "protein_effects", []
        )
        call["annotations"] = variant_interpretation_data.get(
            "annotations", {}
        )
        call[__TYPE__] = f"ANNOTATED_{call.get(__TYPE__, 'CALL')}"

    elif DocType.HEADER.is_a(call):
        call[__CHILD__] = f"ANNOTATED_{call.get(__CHILD__, 'CALL')}"

    return call



from cytoolz.curried import assoc, compose
from genomoncology.parse.doctypes import DocType, __CHILD__, __TYPE__
from genomoncology.pipeline.transformers import (
    register,
    name_mapping,
    get_boolean_value,
)


NAME_MAPPING = {
    "gene": "gene",
    "oncogene__boolean": "oncogene",
    "tumor_suppressor__boolean": "tumor_suppressor",
    "source__string": "source",
    "NCBI_id": "NCBI_id",
}

register(
    input_type=DocType.TSV,
    output_type=DocType.GENE_TYPE,
    transformer=compose(
        lambda x: assoc(
            x,
            "tumor_suppressor__boolean",
            get_boolean_value(x.get("tumor_suppressor__boolean", "")),
        ),
        lambda x: assoc(
            x,
            "oncogene__boolean",
            get_boolean_value(x.get("oncogene__boolean", "")),
        ),
        lambda x: assoc(x, "is_gene_annotation", True),
        lambda x: assoc(x, __TYPE__, DocType.GENE_TYPE.value),
        name_mapping(NAME_MAPPING),
    ),
)

register(
    input_type=DocType.TSV,
    output_type=DocType.GENE_TYPE,
    transformer=compose(
        lambda x: assoc(
            x,
            "tumor_suppressor__boolean",
            get_boolean_value(x.get("tumor_suppressor__boolean", "")),
        ),
        lambda x: assoc(
            x,
            "oncogene__boolean",
            get_boolean_value(x.get("oncogene__boolean", "")),
        ),
        lambda x: assoc(x, "is_gene_annotation", True),
        lambda x: assoc(x, __CHILD__, DocType.GENE_TYPE.value),
    ),
    is_header=True,
)



from .base import BaseInfoMapper, register
from .utils import int_or_none

FAD = "FAD"
FRD = "FRD"
FDP = "FDP"


class BRLIInfoMapper(BaseInfoMapper):
    @classmethod
    def keys(cls):
        return {FAD, FRD, FDP}

    def do_mapping(self, call_info, allele_position, alt_allele):
        total_depth = call_info.get(FDP)

        ref_depth = int_or_none(call_info.get(FRD))

        fad = call_info.get(FAD)
        alt_depth = (
            int_or_none(fad[allele_position - 1])
            if isinstance(fad, (tuple, list))
            else int_or_none(fad)
        )

        return total_depth, alt_depth, ref_depth


register(BRLIInfoMapper)



from cytoolz.curried import assoc, compose
from genomoncology.parse.doctypes import DocType, __CHILD__, __TYPE__
from genomoncology.pipeline.transformers import register, name_mapping

NAME_MAPPING = {
    "gene": "gene",
    "accessions__mstring": "accessions",
    "domains__mstring": "domains",
    "domain_starts__mstring": "domain_starts",
    "domain_ends__mstring": "domain_ends",
}

register(
    input_type=DocType.UNIPROT_RECORD,
    output_type=DocType.UNIPROT_DOMAIN,
    transformer=compose(
        lambda x: assoc(x, "is_gene_annotation", True),
        lambda x: assoc(x, __TYPE__, DocType.UNIPROT_DOMAIN.value),
        name_mapping(NAME_MAPPING),
    ),
)

register(
    input_type=DocType.UNIPROT_RECORD,
    output_type=DocType.UNIPROT_DOMAIN,
    transformer=compose(
        lambda x: assoc(x, "is_gene_annotation", True),
        lambda x: assoc(x, __CHILD__, DocType.UNIPROT_DOMAIN.value),
    ),
    is_header=True,
)



import xml.etree.ElementTree as et

from cytoolz.curried import curry

from .one_off_sources import ClinvarXMLSource


@curry
class XMLSource:
    def __init__(self, file_paths, data_set_name=None):
        self.file_paths = file_paths
        self.data_set_name = data_set_name

    def __iter__(self):
        yield from self.iterate_xml(self.file_paths, self.data_set_name)

    def iterate_xml(self, file_paths, data_set_name):
        assert not isinstance(
            file_paths, str
        ), "Requires list or iterator of str."
        for file_path in file_paths:
            source = self.determine_xml_source(data_set_name, file_path)
            return source

    def get_root(self, file_path):
        try:
            root = et.parse(file_path)
            return root
        except et.ParseError:
            exit()

    def determine_xml_source(self, data_set_name, file_path):
        if data_set_name == "clinvar_xml":
            return ClinvarXMLSource(file_path)
        elif data_set_name == "clinvar_grch38_xml":
            return ClinvarXMLSource(file_path, is_grch38=True)
        return None



import asyncio
from cytoolz.curried import curry, partition_all, concat, map

from gosdk import logger
from genomoncology.parse.ensures import ensure_collection
from genomoncology.kms.annotations import AnnBundleVersionAssertionError


def create_async_processor(
    state: "State", async_func, futures_batch_size=25, **kw
):
    sdk = state.create_sdk(async_enabled=True, timeout=state.timeout)
    return [
        map(async_func(sdk=sdk, **kw)),
        # roll up the futures in batches, await, flatten
        partition_all(futures_batch_size),
        map(do_await(state)),
        concat,
    ]


def create_sync_processor(state: "State", sync_func, **kw):
    sdk = state.create_sdk(async_enabled=False, timeout=state.timeout)
    return map(sync_func(sdk=sdk, **kw))


@curry
def do_await(state, pending):
    """
    pending must be a batch of futures from the async processor
    """
    exceptions_to_always_raise = (AnnBundleVersionAssertionError,)

    pending = ensure_collection(pending)
    logger.get_logger().debug("num pending per do_await: " + str(len(pending)))

    results = state.loop.run_until_complete(
        asyncio.gather(*pending, loop=state.loop, return_exceptions=True)
    )

    for result in results:

        if isinstance(result, exceptions_to_always_raise):
            raise result

        if isinstance(result, Exception):
            state.add_failure(result)
            logger.get_logger().error("do_await failed", error=result)
            if state.hard_failure:
                raise result

        else:
            yield result



from cytoolz.curried import curry

from genomoncology.parse import ensures, DocType, __TYPE__, __CHILD__
from .base import LazyFileSource
import csv


def do_split(delimiter, s):
    s = s.strip()
    if delimiter == ",":
        # reference: https://stackoverflow.com/a/35822856
        val = next(csv.reader([s]))
    else:
        val = s.split(delimiter)
    return val


@curry
class DelimitedFileSource(LazyFileSource):
    def __init__(
        self,
        filename,
        columns,
        delimiter="\t",
        skip_comment=False,
        comment_char="#",
        include_header=True,
        **meta
    ):
        self._mapper = None
        self.columns = ensures.ensure_collection(columns)
        self.delimiter = delimiter
        self.skip_comment = skip_comment
        self.comment_char = comment_char
        self.include_header = include_header
        self.meta = meta

        if __TYPE__ not in meta:
            self.meta[__TYPE__] = DocType.TSV.value

        super().__init__(filename)

    def __iter__(self):
        # noinspection PyUnresolvedReferences
        iterator = super(DelimitedFileSource.func, self).__iter__()
        if self.skip_comment:
            def is_not_comment(line):
                return not line.startswith(self.comment_char)
            iterator = filter(is_not_comment, iterator)

        if not self.columns:
            self.columns = next(csv.reader(iterator, delimiter=self.delimiter))

        if self.include_header:
            yield self.create_header()

        reader = csv.DictReader(iterator, fieldnames=self.columns,
                                delimiter=self.delimiter)
        yield from map(lambda row: {**row, **self.meta}, reader)

    def create_header(self):
        return {
            __TYPE__: DocType.HEADER.value,
            __CHILD__: self.meta.get(__TYPE__),
            "columns": self.columns,
            "meta": self.meta,
            "file_path": self.name,
        }



from genomoncology import kms
from cytoolz.curried import assoc, compose
from genomoncology.parse.doctypes import DocType, __CHILD__, __TYPE__
from genomoncology.pipeline.transformers import register, name_mapping


NAME_MAPPING = {
    # hgvs
    "hgvs_g": "hgvs_g",
    "csra": "csra",
    "chr": "#CHROM",
    "start": "POS",
    "position": "POS",
    "ref": "REF",
    "alt": "ALT",
    "gene": "Gene name",
    "CNT__int": "CNT",
    "tissues__mstring": "TISSUES",
    "tissue_cnt__mstring": "TISSUES",
    "tissue_frequency__mstring": "TISSUES_FREQ",
    "tissue_subtype_cnt__mstring": "TISSUES_SUBTYPE",
    "tissue_subtype_frequency__mstring": "TISSUES_SUBTYPE_FREQ",
    "histology_cnt__mstring": "HISTOLOGY",
    "histology_frequency__mstring": "HISTOLOGY_FREQ",
    "histology_subtype_cnt__mstring": "HISTOLOGY_SUBTYPE",
    "histology_subtype_frequency__mstring": "HISTOLOGY_SUBTYPE_FREQ",
    "resistance_mutation__mstring": "RESISTANCE_MUTATION",
    "genomicID__string": "ID",
    "ID__mstring": "MUTATION_ID",
    "CDS__mstring": "CDS",
    "AA__mstring": "AA",
    "gene_name__mstring": "Gene name",
    "legacy_id__string": "LEGACY_ID",
}


def create_alterations_field(x):
    alterations = set()
    try:
        gene_names = x.get("gene_name__mstring", [])
        aa = x.get("AA__mstring", [])
        cds = x.get("CDS__mstring", [])
        if len(gene_names) != len(aa) != len(cds):
            return None
        # clean gene names
        gene_names = clean_gene(gene_names, make_set=False)
        for i in range(len(gene_names)):
            gene_val = gene_names[i]
            aa_val = aa[i]
            cds_val = cds[i]
            aa_val = aa_val.split("p.")[-1]
            if "=" in aa_val:
                continue
            if aa_val != "?":
                alterations.add(gene_val + " " + aa_val)
            else:
                alterations.add(gene_val + " " + cds_val)

    except IndexError:
        alterations = set()

    return list(alterations)


def get_keys(x, field):
    return list(x.get(field, {}).keys())


def process_frequency_and_counts(x, field):
    freq_results = []
    freq_info = x.get(field)
    for tissue, freq in freq_info.items():
        freq_results.append(f"{tissue}={freq}")
    return freq_results


def clean_gene(gene_list, make_set=True):
    if not isinstance(gene_list, list):
        gene_list = [gene_list]
    # remove everything after "_ENST"
    gene_list = [
        gene.split("_ENST")[0] for gene in gene_list if gene is not None
    ]
    if make_set:
        return list(set(gene_list))
    return gene_list


def remove_start_field(x):
    x.pop("start", "")
    return x


def handle_hgvs_g(x):
    if (
        x.get("chr") is None
        and x.get("start") is None
        and x.get("ref") is None
        and x.get("alt") is None
    ):
        return None
    return kms.annotations.to_csra(x)


register(
    input_type=DocType.AGGREGATE,
    output_type=DocType.COSMIC,
    transformer=compose(
        lambda x: assoc(x, "alterations", create_alterations_field(x)),
        lambda x: assoc(
            x,
            "resistance_mutation__mstring",
            get_keys(x, "resistance_mutation__mstring"),
        ),
        lambda x: assoc(
            x, "tissues__mstring", get_keys(x, "tissues__mstring")
        ),
        lambda x: assoc(
            x,
            "tissue_cnt__mstring",
            process_frequency_and_counts(x, "tissue_cnt__mstring"),
        ),
        lambda x: assoc(
            x,
            "tissue_frequency__mstring",
            process_frequency_and_counts(x, "tissue_frequency__mstring"),
        ),
        lambda x: assoc(
            x,
            "tissue_subtype_cnt__mstring",
            process_frequency_and_counts(x, "tissue_subtype_cnt__mstring"),
        ),
        lambda x: assoc(
            x,
            "tissue_subtype_frequency__mstring",
            process_frequency_and_counts(
                x, "tissue_subtype_frequency__mstring"
            ),
        ),
        lambda x: assoc(
            x,
            "histology_cnt__mstring",
            process_frequency_and_counts(x, "histology_cnt__mstring"),
        ),
        lambda x: assoc(
            x,
            "histology_frequency__mstring",
            process_frequency_and_counts(x, "histology_frequency__mstring"),
        ),
        lambda x: assoc(
            x,
            "histology_subtype_cnt__mstring",
            process_frequency_and_counts(x, "histology_subtype_cnt__mstring"),
        ),
        lambda x: assoc(
            x,
            "histology_subtype_frequency__mstring",
            process_frequency_and_counts(
                x, "histology_subtype_frequency__mstring"
            ),
        ),
        lambda x: remove_start_field(x),
        lambda x: assoc(x, "csra", handle_hgvs_g(x)),
        lambda x: assoc(x, "hgvs_g", handle_hgvs_g(x)),
        lambda x: assoc(x, "gene", clean_gene(x.get("gene", []))),
        lambda x: assoc(x, __TYPE__, DocType.COSMIC.value),
        name_mapping(NAME_MAPPING),
    ),
)

register(
    input_type=DocType.AGGREGATE,
    output_type=DocType.COSMIC,
    transformer=compose(lambda x: assoc(x, __CHILD__, DocType.COSMIC.value)),
    is_header=True,
)



from genomoncology.parse.doctypes import DocType
from genomoncology.pipeline.converters import cint
from genomoncology.pipeline.transformers.registry import register


def wgl_transform(x):
    def tr(v, f=None):
        f = f or int
        v = v.replace("NA", "")

        if v:
            return [f(s.strip()) for s in v.split(",")]
        else:
            return []

    return dict(
        hgvs_c=x.get("hgvs_c"),
        dna_number__mint=tr(x.get("dna_number")),
        parents__mint=tr(x.get("parents")),
        category__mstring=tr(x.get("category"), f=str),
        proband__int=cint(x.get("proband")),
    )


register(output_type=DocType.BAYLOR_WGL, transformer=wgl_transform)



from .base import BaseInfoMapper, register
from .utils import int_or_zero


FAO = "FAO"
FDP = "FDP"
FRO = "FRO"


class FaoFdpInfoMapper(BaseInfoMapper):
    @classmethod
    def keys(cls):
        return {FAO, FDP}  # FRO is optional

    def do_mapping(self, call_info, allele_position, alt_allele):
        total_depth = call_info.get(FDP)

        alt_depth = call_info.get(FAO)
        alt_depth = (
            int_or_zero(alt_depth[allele_position - 1])
            if isinstance(alt_depth, (tuple, list))
            else int_or_zero(alt_depth)
        )

        ref_depth = call_info.get(FRO)
        if ref_depth is None:
            ref_depth = total_depth - alt_depth

        return total_depth, alt_depth, ref_depth


register(FaoFdpInfoMapper)



from cytoolz.curried import assoc, compose
from genomoncology.parse.doctypes import DocType, __CHILD__, __TYPE__
from genomoncology.pipeline.transformers import register, name_mapping


def parse_float(f):
    try:
        return float(f)
    except ValueError:
        return None


def parse_int(i):
    try:
        return int(i)
    except ValueError:
        return None


def parse_values(x):
    # have to do these conversions due to scientific
    # notation and possible "NA" values
    for name, value in x.items():
        if name.endswith("__float"):
            x[name] = parse_float(value)
        elif name.endswith("__int"):
            x[name] = parse_int(value)
    return x


NAME_MAPPING = {
    "gene": "gene",
    "ensembl_gene_id": "gene_id",
    "pLI_score__float": "pLI",
    "mis_z__float": "mis_z",
    "syn_z__float": "syn_z",
    "lof_z__float": "lof_z",
    "oe_lof_upper_rank__int": "oe_lof_upper_rank",
    "oe_lof_upper_bin__int": "oe_lof_upper_bin",
    "oe_lof_upper_bin_6__int": "oe_lof_upper_bin_6",
    "oe_mis__float": "oe_mis",
    "oe_lof__float": "oe_lof",
    "oe_mis_pphen__float": "oe_mis_pphen",
    "oe_syn__float": "oe_syn",
    "oe_syn_lower__float": "oe_syn_lower",
    "oe_syn_upper__float": "oe_syn_upper",
    "oe_mis_lower__float": "oe_mis_lower",
    "oe_mis_upper__float": "oe_mis_upper",
    "oe_lof_lower__float": "oe_lof_lower",
    "oe_lof_upper__float": "oe_lof_upper",
}

register(
    input_type=DocType.TSV,
    output_type=DocType.GNOMAD_GENE,
    transformer=compose(
        lambda x: parse_values(x),
        lambda x: assoc(x, __TYPE__, DocType.GNOMAD_GENE.value),
        lambda x: assoc(x, "is_gene_annotation", True),
        name_mapping(NAME_MAPPING),
    ),
)

register(
    input_type=DocType.TSV,
    output_type=DocType.GNOMAD_GENE,
    transformer=compose(
        lambda x: assoc(x, __CHILD__, DocType.GNOMAD_GENE.value)
    ),
    is_header=True,
)



from cytoolz.curried import assoc, compose
from genomoncology.parse.doctypes import DocType, __CHILD__, __TYPE__
from genomoncology.pipeline.transformers import register, name_mapping


NAME_MAPPING = {
    "gene": "gene",
    # info
    "NCBI_id": "gene_id",
    "phenotype__mstring": "phenotype",
    "hpo__mstring": "hpo_id",
}


def parse_hpo_gene(x):
    return x["key"]


def dedupe_gene_id(x):
    gene_ids = x.get("gene_id", None)
    if gene_ids:
        return list(set(gene_ids))


register(
    input_type=DocType.AGGREGATE,
    output_type=DocType.HPO,
    transformer=compose(
        lambda x: assoc(x, "is_gene_annotation", True),
        lambda x: assoc(x, __TYPE__, DocType.HPO.value),
        name_mapping(NAME_MAPPING),
        lambda x: assoc(x, "gene_id", dedupe_gene_id(x)),
        lambda x: assoc(x, "gene", parse_hpo_gene(x)),
    ),
)

register(
    input_type=DocType.AGGREGATE,
    output_type=DocType.HPO,
    transformer=compose(
        lambda x: assoc(x, "is_gene_annotation", True),
        lambda x: assoc(x, __CHILD__, DocType.HPO.value),
    ),
    is_header=True,
)



INPUT_IS_FILE_PATH_LIST = "__INPUT_IS_FILE_PATH_LIST__"
OUTPUT_IS_EXCEL_FILE = "__OUTPUT_IS_EXCEL_FILE__"
SKIP_CONVERT_TO_JSON = "__SKIP_CONVERT_TO_JSON__"


GRCh37 = "GRCh37"
GRCh38 = "GRCh38"
BUILDS = [GRCh37, GRCh38]



import os
from configparser import ConfigParser

# retrieve variables from environment (higher precedence) or .ini (lower)

CONFIG_SECTION = "gocli-options"
CONFIG_FILE_PATH = "~/.gocli.ini"

# ini

config = ConfigParser()
config.read(os.path.expanduser(CONFIG_FILE_PATH))


def get_variable(variable_name: str):
    return os.getenv(
        f"KMS_{variable_name.upper()}",
        config.get(CONFIG_SECTION, variable_name, fallback=None),
    )



from cytoolz.curried import assoc, compose

from genomoncology.parse.doctypes import DocType, __TYPE__, __CHILD__
from genomoncology.pipeline.transformers import name_mapping
from genomoncology.pipeline.transformers.registry import register

NAME_MAPPING = {"gene": "gene", "hgnc_id": "hgncID"}


register(
    input_type=DocType.TSV,
    output_type=DocType.HGMD_GENE_LIST,
    transformer=compose(
        lambda x: assoc(x, __TYPE__, DocType.HGMD_GENE_LIST.value),
        lambda x: assoc(x, "is_gene_annotation", True),
        lambda x: assoc(x, "is_hgmd_gene__boolean", True),
        name_mapping(NAME_MAPPING),
    ),
    is_header=False,
)

register(
    input_type=DocType.TSV,
    output_type=DocType.HGMD_GENE_LIST,
    transformer=compose(
        lambda x: assoc(x, "is_gene_annotation", True),
        lambda x: assoc(x, __CHILD__, DocType.HGMD_GENE_LIST.value),
    ),
    is_header=True,
)



from functools import partial

from cytoolz.curried import curry

from genomoncology.pipeline.converters import obj_to_dict
from genomoncology.parse.ensures import ensure_collection
from genomoncology.parse.doctypes import DocType
from genomoncology.cli.const import GRCh37


@curry
def sync_genes(data, sdk=None, fields=("chromosome", "start", "end", "name")):
    batch = ensure_collection(data)
    kwargs = dict(name=batch, fields=fields, page_size=len(batch))
    gene_list = sdk.call_with_retry(sdk.genes.get_genes, **kwargs)
    func = partial(obj_to_dict, fields, __type__=DocType.GENE.value)
    return list(map(func, gene_list.results))


@curry
def sync_boundaries(
    data, sdk=None, fields=("chromosome", "start", "end", "gene"), build=GRCh37
):
    batch = ensure_collection(data)
    kwargs = dict(name=batch, build=build)
    gene_list = sdk.call_with_retry(sdk.genes.gene_boundaries, **kwargs)
    func = partial(obj_to_dict, fields, __type__=DocType.GENE.value)
    formatted_results = list(map(func, gene_list.results))
    if gene_list.not_found:
        message = "The following genes could not be found: " + ", ".join(
            gene_list.not_found
        )
        formatted_results.append(message)
    return formatted_results



from cytoolz.curried import curry

from gosdk import logger


@curry
def process_transcript_batch(data, sdk=None):
    record = {
        "batch": [
            "%s|%s|%s" % (record["chromosome"], record["start"], record["end"])
            for record in data
        ]
    }
    logger.get_logger().debug("get_transcripts_batch", **record)

    results = sdk.call_with_retry(
        sdk.region_search.region_search_batch, **record
    )["results"]
    return results


@curry
def process_transcript_batch_get_genes(data, sdk=None):
    record = {
        "batch": [
            "%s|%s|%s" % (record["chromosome"], record["start"], record["end"])
            for record in data
        ]
    }
    logger.get_logger().debug("get_transcripts_batch", **record)
    # Get all regions for the genes
    results = sdk.call_with_retry(
        sdk.region_search.region_search_batch, **record
    )["results"]
    # we only need the gene names from the result that is returned,
    # We also do not want duplicate names.
    list_of_genes = []
    for line in results:
        transcripts = line.get("transcripts", [])
        for transcript in transcripts:
            list_of_genes.append(transcript.get("gene"))
    return set(list_of_genes)



# common.py - models that are shared across versions.

from specd.sdk import BaseModel


class AnnotationList(BaseModel):

    # Set of keys that are NOT annotations. New top-level keys must be added.
    META_KEYS = {"errors", "is_created", "additionalProperties", "normalized"}

    def get_annotation(self, key):
        obj_dict = getattr(self, key, None)
        if obj_dict:
            return self.instantiate(self.definitions.Annotation, obj_dict)

    def iter_annotations(self):
        for key in self:
            if key not in AnnotationList.META_KEYS:
                yield self.get_annotation(key)


class AnnotationMatchList(BaseModel):
    # Set of keys that are NOT annotations. New top-level keys must be added.
    META_KEYS = {"errors", "is_created", "additionalProperties", "normalized"}

    def get_annotation(self, key):
        obj_dict = getattr(self, key, None)
        if obj_dict:
            return self.instantiate(self.definitions.AnnotationMatch, obj_dict)

    def iter_annotations(self):
        for key in self:
            if key not in AnnotationMatchList.META_KEYS:
                yield self.get_annotation(key)


class VariantInterpretationResponse(BaseModel):

    # set of keys that are NOT part of the variant interpretations response
    META_KEYS = {"errors", "additionalProperties"}

    def get_var_int_data(self, key):
        obj_dict = getattr(self, key, None)
        if obj_dict:
            return self.instantiate(
                self.definitions.VariantInterpretationSingleResponse, obj_dict
            )

    def iter_var_int_data(self):
        for key in self:
            if key not in VariantInterpretationResponse.META_KEYS:
                yield self.get_var_int_data(key)


# MEGA-MATCH MODELS


def convert_object_to_dict(o):
    if o is None:
        return o
    return {key: getattr(o, key) for key in o}


class MegaMatchSingleTherapyObject(BaseModel):
    def convert_to_dict(self):
        d = {}
        for key in self:
            if key in [
                "detected_alterations",
                "match_results",
                "matched_diseases",
            ]:
                d[key] = [
                    convert_object_to_dict(o) for o in getattr(self, key, [])
                ]
            else:
                d[key] = getattr(self, key)
        return d


class MegaMatchTherapyDocument(BaseModel):
    def convert_single_therapy_to_dict(self, t: MegaMatchSingleTherapyObject):
        if t:
            t = t.convert_to_dict()
        return t

    def convert_to_dict(self):
        d = {}
        for key in self:
            if key == "consolidated":
                d[key] = self.convert_single_therapy_to_dict(
                    getattr(self, key)
                )
            elif key in [
                "approvals",
                "guidelines",
                "special_statuses",
                "evidence",
                "associations",
            ]:
                therapies = getattr(self, key, [])
                d[key] = [
                    self.convert_single_therapy_to_dict(t) for t in therapies
                ]
            else:
                d[key] = getattr(self, key, None)
        return d


class MegaMatchTrialDocument(BaseModel):
    def convert_to_dict(self):
        d = {}
        for key in self:
            if key == "clinical_trial_arms":
                arms = getattr(self, key, [])
                d[key] = [convert_object_to_dict(arm) for arm in arms]
            else:
                d[key] = getattr(self, key, None)
        return d


class MegaMatchResult(BaseModel):
    def convert_therapy_to_dict(self, t: MegaMatchTherapyDocument):
        if t:
            t = t.convert_to_dict()
        return t

    def convert_trial_to_dict(self, t: MegaMatchTrialDocument):
        if t:
            t = t.convert_to_dict()
        return t

    def convert_to_dict(self):
        # convert the therapies/clinical_trials fields from lists of model
        # instances to lists of dicts
        d = {}
        for key in self:
            if key == "therapies":
                therapies = getattr(self, key, [])
                d[key] = [self.convert_therapy_to_dict(t) for t in therapies]
            elif key == "clinical_trials":
                trials = getattr(self, key, [])
                d["clinical_trials"] = [
                    self.convert_trial_to_dict(t) for t in trials
                ]
            else:
                d[key] = getattr(self, key, None)
        return d



from genomoncology import kms
from cytoolz.curried import assoc, compose
from genomoncology.parse.doctypes import DocType, __CHILD__, __TYPE__
from genomoncology.pipeline.transformers import register, name_mapping


NAME_MAPPING = {
    # hgvs
    "chr": "chr",
    "start": "start",
    "ref": "ref",
    "alt": "alt",
    "build": "build",
    # info
    "AC__int": "info.AC",
    "AC_Adj__int": "info.AC_Adj",
    "AF__float": "info.AF",
    "AN__int": "info.AN",
    "AN_Adj__int": "info.AN_Adj",
    "AC_AFR__int": "info.AC_AFR",
    "AC_AMR__int": "info.AC_AMR",
    "AC_EAS__int": "info.AC_EAS",
    "AC_FIN__int": "info.AC_FIN",
    "AC_NFE__int": "info.AC_NFE",
    "AC_OTH__int": "info.AC_OTH",
    "AC_SAS__int": "info.AC_SAS",
    "AN_AFR__int": "info.AN_AFR",
    "AN_AMR__int": "info.AN_AMR",
    "AN_EAS__int": "info.AN_EAS",
    "AN_FIN__int": "info.AN_FIN",
    "AN_NFE__int": "info.AN_NFE",
    "AN_OTH__int": "info.AN_OTH",
    "AN_SAS__int": "info.AN_SAS",
    "Hemi_AFR__int": "info.Hemi_AFR",
    "Hemi_AMR__int": "info.Hemi_AMR",
    "Hemi_EAS__int": "info.Hemi_EAS",
    "Hemi_FIN__int": "info.Hemi_FIN",
    "Hemi_NFE__int": "info.Hemi_NFE",
    "Hemi_OTH__int": "info.Hemi_OTH",
    "Hemi_SAS__int": "info.Hemi_SAS",
    "Het_AFR__mint": "info.Het_AFR",
    "Het_AMR__mint": "info.Het_AMR",
    "Het_EAS__mint": "info.Het_EAS",
    "Het_FIN__mint": "info.Het_FIN",
    "Het_NFE__mint": "info.Het_NFE",
    "Het_OTH__mint": "info.Het_OTH",
    "Het_SAS__mint": "info.Het_SAS",
    "Hom_AFR__int": "info.Hom_AFR",
    "Hom_AMR__int": "info.Hom_AMR",
    "Hom_EAS__int": "info.Hom_EAS",
    "Hom_FIN__int": "info.Hom_FIN",
    "Hom_NFE__int": "info.Hom_NFE",
    "Hom_OTH__int": "info.Hom_OTH",
    "Hom_SAS__int": "info.Hom_SAS",
}

register(
    input_type=DocType.CALL,
    output_type=DocType.EXAC,
    transformer=compose(
        lambda x: assoc(x, "hgvs_g", kms.annotations.to_csra(x)),
        lambda x: assoc(x, __TYPE__, DocType.EXAC.value),
        name_mapping(NAME_MAPPING),
    ),
)

register(
    input_type=DocType.CALL,
    output_type=DocType.EXAC,
    transformer=compose(lambda x: assoc(x, __CHILD__, DocType.EXAC.value)),
    is_header=True,
)



# need to import all of the mappers so they get registered.

from . import ao_ro_dp
from . import ad_rd_dp
from . import brli
from . import clcbio
from . import fao_fdp
from . import pheo
from . import strelka
from . import usc
from .base import find

__all__ = (
    "strelka",
    "pheo",
    "clcbio",
    "fao_fdp",
    "ao_ro_dp",
    "ad_rd_dp",
    "brli",
    "usc",
    "find",
)



from glom import glom
from cytoolz.curried import curry
from genomoncology.parse.ensures import ensure_collection
from .converters import fint


@curry
def cmp(x, y):
    """
    Replacement for built-in function cmp that was removed in Python 3

    Compare the two objects x and y and return an integer according to
    the outcome. The return value is negative if x < y, zero if x == y
    and strictly positive if x > y.
    """
    return (x > y) - (x < y)


def safe_lower(val):
    val = "|".join(map(str, ensure_collection(val)))
    return val.lower()


@curry
def compare_contains(glom_path, cmp_value, obj):
    # todo: need proper unit tests. currently only supports string contains
    # case insensitive: the string "Likely benign" does contain "Benign"
    obj_value = glom(obj, glom_path, default=None)

    if obj_value is not None:
        return safe_lower(cmp_value) in safe_lower(obj_value)


@curry
def compare_in(glom_path: str, s: set, obj):
    obj_value = glom(obj, glom_path, default=None)
    return s.intersection(ensure_collection(obj_value))


@curry
def compare(glom_path: str, cmp_set: set, cmp_value, obj):
    cmp_value = None if cmp_value == "." else cmp_value
    obj_value = glom(obj, glom_path, default=None)

    # todo: this isn't correct, using first item as representative
    # todo: switch to a loop and return True if ANY (?)
    if obj_value and isinstance(obj_value, (list,)):
        obj_value = obj_value[0]

    if isinstance(obj_value, (int, float)):
        cmp_value = fint(cmp_value)

    if None in {obj_value, cmp_value}:
        return obj_value is None and cmp_value is None
    else:
        return cmp(obj_value, cmp_value) in cmp_set


LT = {-1}
LTE = {-1, 0}
GT = {1}
GTE = {1, 0}
EQ = {0}
NE = {1, -1}

COMPARE_LOOKUP = {
    "<": LT,
    ">": GT,
    ">=": GTE,
    "<=": LTE,
    "==": EQ,
    "=": EQ,
    "!=": NE,
    "<>": NE,
    "LT": LT,
    "GT": GT,
    "GTE": GTE,
    "LTE": LTE,
    "EQ": EQ,
    "NE": NE,
}



from cytoolz.curried import assoc, compose
from genomoncology.parse.doctypes import DocType, __CHILD__, __TYPE__
from genomoncology.pipeline.transformers import register, name_mapping
from genomoncology.pipeline.transformers.functions import turn_gene_to_list

NAME_MAPPING = {
    "gene": "gene",
    "disease_name__string": "disease_name",
    "disease_mim__string": "disease_mim",
    "gene_mim": "MIM",
}


register(
    input_type=DocType.TSV,
    output_type=DocType.SECONDARY_FINDINGS_GENES,
    transformer=compose(
        lambda x: assoc(x, "is_gene_annotation", True),
        lambda x: assoc(x, __TYPE__, DocType.SECONDARY_FINDINGS_GENES.value),
        lambda x: assoc(x, "gene", turn_gene_to_list(x)),
        name_mapping(NAME_MAPPING),
    ),
)

register(
    input_type=DocType.TSV,
    output_type=DocType.SECONDARY_FINDINGS_GENES,
    transformer=compose(
        lambda x: assoc(x, "is_gene_annotation", True),
        lambda x: assoc(x, __CHILD__, DocType.SECONDARY_FINDINGS_GENES.value),
    ),
    is_header=True,
)



from .construct import construct_sdk
from .variables import get_variable
from . import logger, models

__all__ = ("construct_sdk", "get_variable", "logger", "models")

__author__ = """Ian Maurer"""
__email__ = "ian@genomoncology.com"
__version__ = "0.9.23"

__uri__ = "http://pypi.org/project/gosdk"
__copyright__ = "Copyright (c) 2018 genomoncology.com"
__description__ = "govcf: GenomOncology Software Development Kit"
__doc__ = __description__ + " <" + __uri__ + ">"
__title__ = "gosdk"



import asyncio
import time

import related

from gosdk import construct_sdk


@related.mutable()
class State(object):
    debug = related.BooleanField(default=False)
    quiet = related.BooleanField(default=False)
    hard_failure = related.BooleanField(default=True)
    start_time = related.IntegerField(required=False)
    end_time = related.IntegerField(required=False)
    glob = related.SequenceField(str, required=False, repr=True)
    include_tar = related.BooleanField(default=False)

    batch_size = related.IntegerField(default=50)
    timeout = related.IntegerField(default=300)

    loop = related.ChildField(object, required=False, repr=False)
    run_id = related.StringField(required=False)
    pipeline = related.StringField(required=False)
    build = related.StringField(required=False)
    sdk_refs = related.SequenceField(object, default=[])
    failures = related.SequenceField(object, default=[])
    token = related.StringField(required=False)
    host = related.StringField(required=False)
    schemes = related.SequenceField(str, required=False)
    max_time = related.IntegerField(required=False)

    def set_start(self):
        self.start_time = time.time()

    def set_end(self):
        self.end_time = time.time()

    def create_sdk(self, async_enabled=False, loop=None, timeout=300):
        sdk = construct_sdk(
            async_enabled=async_enabled,
            loop=loop or self.loop,
            token=self.token,
            host=self.host,
            schemes=self.schemes,
            timeout=timeout,
            custom_header_option=self.custom_header_option,
        )
        if async_enabled:
            self.sdk_refs.append(sdk)
        return sdk

    def add_failure(self, failure):
        self.failures.append(failure)

    def finalize(self):
        pending = []

        for sdk_ref in self.sdk_refs:
            pending.append(sdk_ref.close())

        if pending:
            self.loop.run_until_complete(asyncio.wait(pending))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup

with open('./src/genomoncology/README.md') as readme_file:
    readme = readme_file.read()

history = ""

tests_require = [
    "flake8 == 6.1.0",
    "black == 23.10.1",
    "ipython == 8.17.2",
    "pytest == 7.4.3",
    "pytest-cov == 4.1.0",
    "white == 0.1.2",
    "pytest-socket == 0.6.0",
    "pytest-asyncio == 0.21.1",
    "aioconsole == 0.6.2",
    "addict == 2.4.0",
    "requests >= 2.31.0",
    "twine == 4.0.2",
    "readme-renderer == 42.0"
]

setup(
    name="gocli",
    version='0.10.1',
    author="Ian Maurer",
    author_email='ian@genomoncology.com',

    packages=[
        "genomoncology",
        "genomoncology.kms",
        "genomoncology.cli",
        "genomoncology.parse",
        "genomoncology.pipeline",
        "genomoncology.pipeline.sinks",
        "genomoncology.pipeline.sources",
        "genomoncology.pipeline.sources.one_off_sources",
        "genomoncology.pipeline.transformers",
        "genomoncology.pipeline.transformers.tx",
        "gosdk",
        "govcf",
        "govcf.calculate_vaf",

    ],
    package_dir={
        '': 'src'
    },

    package_data={
        '': ["*.yaml", "*.bed", "*.txt", "*.tsv", "*.csv"]
    },

    include_package_data=True,

    tests_require=tests_require,
    install_requires=[
        "related >= 0.7.3",
        "backoff == 2.2.1",
        "click == 8.1.7",
        "structlog == 23.2.0",
        "colorama == 0.4.6",
        "pysam == 0.22.0",
        "ujson == 5.8.0",
        "intervaltree == 3.1.0",
        "glom == 23.3.0",
        "cytoolz == 0.12.2",
        "openpyxl == 3.1.2",
        "pygments == 2.16.1",
        "jsonschema[format] == 4.19.2",
        "flask == 1.0.2",
        "itsdangerous == 2.1.1",
        "Werkzeug == 3.0.1",
        "jinja2 == 3.1.2",
        "flask-swagger-ui == 4.11.1",
        "swagger-spec-validator == 3.0.3",
        "PyYAML == 6.0.1",
        "dictdiffer == 0.9.0",
        "inflect == 7.0.0",
        "aiobravado == 0.9.2",
        "bravado == 11.0.3",
        "bravado-core == 6.1.0",
        "bravado-asyncio == 2.0.2",
        "stringcase == 1.2.0",
        "urllib3 == 2.0.7",
    ],

    setup_requires=[
        'pytest-runner',
    ],

    license="Proprietary",
    keywords='Bioinformatics HGVS VCF Clinical Trials Genomics',

    description="gocli",
    long_description="%s\n\n%s" % (readme, history),

    entry_points={
        'console_scripts': [
            'gocli=genomoncology.main:main',
        ],
    },

    classifiers=[
        'License :: Other/Proprietary License',
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.11',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
    ],
)



import os

import backoff
import specd.sdk
from configparser import ConfigParser
from bravado.exception import HTTPInternalServerError
from urllib3.exceptions import TimeoutError
from aiohttp.client_exceptions import ClientConnectionError

from . import logger, models

# const

KMS_TOKEN = os.getenv("KMS_TOKEN")
KMS_HOST = os.getenv("KMS_HOST")
KMS_SCHEMES = os.getenv("KMS_SCHEMES")
KMS_SCHEMES = KMS_SCHEMES.split(",") if KMS_SCHEMES else None
DEFAULT_SPECD_PATH = os.path.join(os.path.dirname(__file__), "./specs/")
HTTP_EXCEPTIONS = (
    OSError,
    ConnectionError,
    TimeoutError,
    ClientConnectionError,
    # HTTPGatewayTimeout,
)
MAX_TIME = 600


# retrieve variables from environment (higher precedence) or .ini (lower)

CONFIG_SECTION = "gocli-options"
CONFIG_FILE_PATH = "~/.gocli.ini"

# ini

config = ConfigParser()
config.read(os.path.expanduser(CONFIG_FILE_PATH))


def construct_sdk(  # pragma: no mccabe
    targets=None,
    host=None,
    token=None,
    async_enabled=False,
    verify_ssl=False,
    loop=None,
    specd_path=None,
    schemes=None,
    enable_validation=False,
    config=None,
    max_time=MAX_TIME,
    max_tries=12,
    timeout=600,
    custom_header_option=None,
):

    token = token or KMS_TOKEN

    # add custom headers to the current headers var
    # creates a dict of key:val by iterating through
    # the tuple of custom_header_option and initializes
    # the dict by separating each element in the tuple
    # by ':'. ex. ('key1:val1','key2:val2') will be
    # {'key1':'val1','key2:'val2'}

    headers = dict(map(lambda x: x.split(":"), custom_header_option or []))
    if token:
        headers["Authorization"] = f"Token {token}"

    specd_path = specd_path or DEFAULT_SPECD_PATH

    config = config or {}

    # validating requests/responses causes problems for extra fields
    # when they cannot be supported by additionalProperties
    # (e.g. differing types in the same record: ao__int, sig__string)
    if not enable_validation:
        config.update(
            {
                "validate_responses": False,
                "validate_requests": False,
                "validate_swagger_spec": False,
            }
        )

    sdk = specd.sdk.create_sdk(
        specd_path,
        host=host or KMS_HOST,
        targets=targets,
        headers=headers,
        models=models,
        async_enabled=async_enabled,
        verify_ssl=verify_ssl,
        loop=loop,
        schemes=schemes or KMS_SCHEMES,
        config=config,
    )

    if async_enabled:

        @backoff.on_exception(
            backoff.expo,
            HTTP_EXCEPTIONS,
            max_time=max_time,
            max_tries=max_tries,
        )
        async def _async_with_retry(sdk_func, **kwargs):
            try:
                return handle_errors(
                    await sdk_func(**kwargs).result(timeout=timeout)
                )

            except HTTPInternalServerError as error:
                logger.get_logger().error(
                    "_async_with_retry",
                    sdk_func=sdk_func,
                    kwargs=kwargs,
                    error=error,
                )
                raise

        sdk.call_with_retry = _async_with_retry

    else:

        @backoff.on_exception(
            backoff.expo,
            HTTP_EXCEPTIONS,
            max_time=max_time,
            max_tries=max_tries,
        )
        def _sync_with_retry(sdk_func, **kwargs):
            try:
                return handle_errors(
                    sdk_func(**kwargs).result(timeout=timeout)
                )

            except HTTPInternalServerError as error:
                logger.get_logger().error(
                    "_sync_with_retry",
                    sdk_func=sdk_func,
                    kwargs=kwargs,
                    error=error,
                )
                raise

        sdk.call_with_retry = _sync_with_retry

    return sdk


def handle_errors(results):  # pragma: no mccabe
    try:
        errors = results.errors or []
    except Exception:
        errors = []

    for error in errors:
        if isinstance(error, list):
            error = error[0]
        if not error:
            continue
        err_msg = "no message provided."
        if "error_message" in error.keys():
            err_msg = error["error_message"]
        if "error_type" in error.keys():
            err_msg += " - " + error["error_type"]

        logger.get_logger().error("add_batch", error=err_msg)

    return results


def get_variable(variable_name: str):
    return os.getenv(
        f"KMS_{variable_name.upper()}",
        config.get(CONFIG_SECTION, variable_name, fallback=None),
    )



from typing import Union, Iterable, Callable

from cytoolz.curried import map, curry, valfilter

from genomoncology.parse import ensures
from genomoncology.parse.doctypes import DocType, __TYPE__, __CHILD__

SPEC_REGISTRY = {}


def register(
    transformer: Callable,
    output_type: Union[Iterable[DocType], DocType],
    input_type: Union[Iterable[DocType], DocType, None] = None,
    is_header: bool = False,
):

    for out_item in ensures.ensure_collection(output_type):
        # register glom/cytoolz transformation map (see ./specs.py)
        if input_type:
            for in_item in ensures.ensure_collection(input_type):
                SPEC_REGISTRY[(in_item, out_item, is_header)] = transformer

        # register callable catch-all class (e.g. VariantMapper)
        else:
            SPEC_REGISTRY[out_item] = transformer


def get_transformer(
    output_type: DocType,
    input_type: Union[DocType, str] = None,
    is_header: bool = False,
) -> Callable:

    transformer = None

    if input_type:
        input_type = DocType(input_type)
        key = (input_type, output_type, is_header)
        transformer = SPEC_REGISTRY.get(key)

    transformer = transformer or SPEC_REGISTRY.get(output_type)

    return transformer


@curry
def do_transform_to(output_type: DocType, value: dict):
    input_type = value.get(__TYPE__)
    is_header = DocType.HEADER.value == input_type
    if is_header:
        input_type = value.get(__CHILD__)

    transformer = get_transformer(
        output_type=output_type, input_type=input_type, is_header=is_header
    )

    value = transformer(value) if transformer else value
    return value


def create_transformer(output_type: str, state):
    output_type = DocType.create(output_type)

    transformer_cls = get_transformer(output_type)

    # handle class-based transformers (see mapper_classes.py)
    if transformer_cls and isinstance(transformer_cls, type):
        meta = valfilter(
            lambda x: x is not None,
            dict(run_id=state.run_id, pipeline=state.pipeline),
        )
        return map(transformer_cls(**meta))

    # handle functional (i.e. glom, cytoolz) based (see specs.py)
    else:
        return map(do_transform_to(output_type))
